/* $Id: e2_view_dialog.c 1514 2006-01-24 12:46:13Z tpgww $
 * 
 * Copyright (C) 2004-2005 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "emelfm2.h"
#include <string.h>
#include "e2_dialog.h"
#include "e2_view_dialog.h"
#include "e2_textiter.h"

static GList *history = NULL;
static E2_OptionSet *opt_show_last;

static E2_OptionSet *opt_view_width;
static E2_OptionSet *opt_view_height;
static E2_OptionSet *opt_view_wrap;
static E2_ViewDialogRuntime *vrt;

static void _e2_view_dialog_destroy (E2_ViewDialogRuntime *rt);
static gboolean _e2_view_dialog_refresh (GtkWidget *textview, gchar *filename);

static void _e2_view_dialog_update_search_combo (E2_ViewDialogRuntime *rt)
{
	GtkWidget *entry = GTK_BIN (rt->combo)->child;
	if (opt_show_last)
	{
		GtkTreeIter iter;
		gchar *text = e2_tree_get_last (gtk_combo_box_get_model
			(GTK_COMBO_BOX (rt->combo)), 0, &iter);
		gchar *new_text = NULL;
		new_text = g_strdup (text);
		if ((text != NULL) && (g_str_equal (text, new_text)))
			e2_combobox_set_active (rt->combo, e2_tree_count (gtk_combo_box_get_model
				(GTK_COMBO_BOX (rt->combo))) - 1);
		g_free (text);
		if (new_text == NULL)
			gtk_entry_set_text (GTK_ENTRY (entry), "");
		else
		{
			gtk_entry_set_text (GTK_ENTRY (entry), new_text);
			g_free (new_text);
		}
		gtk_editable_select_region (GTK_EDITABLE (entry), 0, -1);
	} else
		gtk_entry_set_text (GTK_ENTRY (entry), "");
}

static void _e2_view_dialog_attach_search_iters (GtkTextBuffer *buffer,
	GtkTextIter *start, GtkTextIter *end)
{
	GtkTextMark *mark = gtk_text_buffer_get_mark (buffer, "e2-search-start");
	if (mark == NULL)
		gtk_text_buffer_create_mark (buffer, "e2-search-start", start, FALSE);
	else
		gtk_text_buffer_move_mark (buffer, mark, start);
	mark = gtk_text_buffer_get_mark (buffer, "e2-search-end");
	if (mark == NULL)
		gtk_text_buffer_create_mark (buffer, "e2-search-end", end, FALSE);
	else
		gtk_text_buffer_move_mark (buffer, mark, end);
}

static void _e2_view_dialog_unattach_search_iters (GtkTextBuffer *buffer)
{
	GtkTextMark *mark = gtk_text_buffer_get_mark (buffer, "e2-search-start");
	if (mark != NULL)
		gtk_text_buffer_delete_mark (buffer, mark);
	mark = gtk_text_buffer_get_mark (buffer, "e2-search-end");
	if (mark != NULL)
		gtk_text_buffer_delete_mark (buffer, mark);
}

static gboolean _e2_view_dialog_dosearch (GtkTextIter *search_iter,
	GtkTextIter *start, GtkTextIter *end, const gchar *find,
	E2_ViewDialogRuntime *rt)
{
	GtkSourceSearchFlags flags = GTK_SOURCE_SEARCH_TEXT_ONLY;
	if (!rt->case_sensitive)
		flags |= GTK_SOURCE_SEARCH_CASE_INSENSITIVE;
	if (rt->search_backward)
		return e2_iter_backward_search (search_iter, find,
			flags, start, end, NULL);
	else
		return e2_iter_forward_search (search_iter, find,
			flags, start, end, NULL);
}

static void _e2_view_dialog_search (gint response, E2_ViewDialogRuntime *rt)
{
	// get the find string
	GtkWidget *entry = GTK_BIN (rt->combo)->child;
	const gchar *find = gtk_entry_get_text (GTK_ENTRY (entry));
	if (*find == '\0')
		return;
	e2_combobox_activated_cb (entry, GINT_TO_POINTER (FALSE));
	e2_list_update_history (find, &history, NULL, 30, FALSE);
	_e2_view_dialog_update_search_combo (rt);
	
	//check if the TextView's buffer has changed and get new TextIters
	GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (rt->textview));
/*	if (buffer != rt->textbuffer)
	{
		rt->textbuffer = buffer;
		unattach_iters (buffer);
		search_iter = NULL;
	}
*/
	GtkTextMark *mark = NULL;
	if (response == E2_RESPONSE_FIND)
		_e2_view_dialog_unattach_search_iters (buffer);	//(re)start searching
	else	//find next
	{
		if (rt->search_backward)
			mark = gtk_text_buffer_get_mark (buffer, "e2-search-start");
		else
			mark = gtk_text_buffer_get_mark (buffer, "e2-search-end");
	}

	GtkTextIter *search_iter = NULL;
	GtkTextIter iter;
	if (mark != NULL)
	{
		gtk_text_buffer_get_iter_at_mark (buffer, &iter, mark);
		search_iter = &iter;
	}
	if (search_iter == NULL)
	{
		//search from cursor
		mark = gtk_text_buffer_get_mark (buffer, "insert");
		gtk_text_buffer_get_iter_at_mark (buffer, &iter, mark);
		search_iter = &iter;
/*	FIXME search from top/bottom of window if cursor is off-window
		GdkRectangle visible;
		gtk_text_view_get_visible_rect (GTK_TEXT_VIEW (rt->textview),
			&visible);
		if (rt->opt_search_backward)
			//search backward from end of screen
			gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (rt->textview),
				&iter, visible.width, visible.height);
		else
			//search forward from top of screen
			gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (rt->textview),
				&iter, visible.x, visible.y); */
	}

	// search for the string
	GtkTextIter start, end;
	gboolean found = _e2_view_dialog_dosearch (search_iter, &start, &end, find, rt);

	if (!found && rt->searchwrap)
	{
		_e2_view_dialog_unattach_search_iters (buffer);
		GtkTextIter iter;
		if (rt->search_backward)
			gtk_text_buffer_get_end_iter (buffer, &iter);
		else
			gtk_text_buffer_get_start_iter (buffer, &iter);
		found = _e2_view_dialog_dosearch (&iter, &start, &end, find, rt);
	}
	if (found)
	{
		gtk_text_buffer_select_range (buffer, &start, &end);
		_e2_view_dialog_attach_search_iters (buffer, &start, &end);
		gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (rt->textview), &start,
			0.2, TRUE, 0.2, 0.5);
		gtk_label_set_text (GTK_LABEL (rt->info_label), "");
	}
	else
	{
		gchar *labeltext = g_strconcat ("<span weight=\"bold\" foreground=\"",
			e2_option_str_get ("color-negative"), "\">", _("not found"), "</span>", NULL);
		gtk_label_set_markup (GTK_LABEL (rt->info_label), labeltext);
		g_free (labeltext);
	}
}

  /*********************/
 /***** callbacks *****/
/*********************/

static void  _e2_view_dialog_search_activate_cb 
	(GtkWidget *entry, E2_ViewDialogRuntime *rt)
{
	_e2_view_dialog_search (E2_RESPONSE_FIND, rt);
}

static void _e2_view_dialog_response_cb
	(GtkWidget *dialog, gint response, E2_ViewDialogRuntime *rt)
{
	printd (DEBUG, "response_cb (dialog:_,response:%d,rt:_)", response);
	switch (response)
	{
		case E2_RESPONSE_USER1:  //text wrap
			rt->textwrap = !rt->textwrap;
			gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (rt->textview),
				rt->textwrap ? GTK_WRAP_WORD: GTK_WRAP_NONE);
			break;
		case E2_RESPONSE_FIND:
			if (GTK_WIDGET_VISIBLE (rt->panel)) //search panel is visible
			{
				_e2_view_dialog_search (E2_RESPONSE_FIND, rt);
			}
			else
			{	//show the search panel
				gtk_widget_show (rt->panel);
				//show extra buttons
				gtk_widget_show (vrt->nextbtn);
				gtk_widget_show (vrt->hidebtn);
				gtk_widget_show (vrt->info_label);
				gtk_widget_hide (vrt->refreshbtn);
			}
			break;
		case E2_RESPONSE_USER2:	//find next
			_e2_view_dialog_search (E2_RESPONSE_USER2, rt);
			break;
		case E2_RESPONSE_USER3:	//hide
			gtk_widget_hide (rt->panel);
			gtk_widget_hide (vrt->nextbtn);
			gtk_widget_hide (vrt->hidebtn);
			gtk_widget_hide (vrt->info_label);
			gtk_widget_show (vrt->refreshbtn);
			break;
		case E2_RESPONSE_REFRESH:
			_e2_view_dialog_refresh (rt->textview, rt->filename);
			break;
		default:
			_e2_view_dialog_destroy (rt);
			return;
			break;
	}
	gtk_widget_grab_focus (GTK_WIDGET (rt->textview));
}

static void  _e2_view_dialog_toggled (GtkWidget *button, gboolean *store)
{
	*store = GTK_TOGGLE_BUTTON (button)->active;
}

  /*****************/
 /***** utils *****/
/*****************/

static void _e2_view_dialog_destroy (E2_ViewDialogRuntime *rt)
{
	gtk_widget_destroy (rt->dialog);
	g_free (rt->filename);
	if (rt->idle_id != 0)
		g_source_remove (rt->idle_id);
	g_free (rt);
}

static gboolean _e2_view_dialog_refresh (GtkWidget *textview, gchar *filepath)
{
	gchar *cont;
	if (! e2_fs_get_file_contents (filepath, &cont))
	{
		printd (WARN, "could not open file '%s' for reading", filepath);
		gchar *msg = g_strdup_printf (_("Could not open file '%s' for reading"), filepath);
		e2_output_print_error (msg);
		g_free (msg);
		return FALSE;
	}
	GtkTextBuffer *buffer = vrt->textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));

	if (!g_utf8_validate (cont, -1, NULL))
	{
		printd (DEBUG, "trying to convert '%s' to UTF-8", filepath);
		GError *utf_error = NULL;
		gchar *utf = g_locale_to_utf8 (cont, -1, NULL, NULL, &utf_error);
		if (utf != NULL)
			gtk_text_buffer_set_text (buffer, utf, -1);
		else
		{
			gtk_text_buffer_set_text (buffer, utf_error->message, -1);
			g_error_free (utf_error);
		}
		g_free (utf);
	}
	else
	{
		printd (DEBUG, "'%s' is UTF-8 encoded", filepath);
		gtk_text_buffer_set_text (buffer, cont, -1);
	}
	g_free (cont);
	//set cursor to start of buffer, for searching etc
	GtkTextIter start;
	gtk_text_buffer_get_start_iter (buffer, &start);
	gtk_text_buffer_place_cursor (buffer, &start);

	return TRUE;
}

static GtkWidget *_e2_view_dialog_create (gchar *filename)
{
	gchar *file;
	if (!g_path_is_absolute (filename))
		file = g_build_filename (curr_view->dir, filename, NULL);
	else
		file = g_strdup (filename);
	gchar *local = F_FILENAME_TO_LOCALE (file);
	if (e2_fs_access (local, F_OK))
	{
		gchar *msg = g_strdup_printf (_("Cannot open '%s', because it does not exist"), file);
		e2_output_print_error (msg);
		g_free (msg);
		g_free (file);
		F_FREE (local);
		return NULL;
	}
	if (e2_fs_is_dir3 (local))
	{
		gchar *msg = g_strdup_printf (_("Cannot view directory '%s' in this file viewer"), file);
		e2_output_print_error (msg);
		g_free (msg);
		g_free (file);
		F_FREE (local);
		return NULL;
	}
	F_FREE (local);
	//init view runtime object
//	E2_ViewDialogRuntime * use the global pointer
	vrt = g_malloc (sizeof (E2_ViewDialogRuntime));
	vrt->filename = file;
	vrt->textwrap = _e2_option_bool_get (opt_view_wrap);
	vrt->window_width = _e2_option_int_get (opt_view_width);
	vrt->window_height = _e2_option_int_get (opt_view_height);
	vrt->idle_id = 0;

	vrt->dialog = e2_dialog_create (NULL, file, _("displaying file"),
		_e2_view_dialog_response_cb, vrt);
	//create the view
	GtkWidget *sw = e2_widget_add_sw (GTK_DIALOG (vrt->dialog)->vbox,
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC, TRUE, E2_PADDING);
	vrt->textview = gtk_text_view_new ();
	gtk_container_add (GTK_CONTAINER (sw), vrt->textview);
	gtk_widget_show (vrt->textview);
	// set view defaults
	gtk_text_view_set_editable (GTK_TEXT_VIEW (vrt->textview), FALSE);
	gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (vrt->textview), FALSE);
	gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (vrt->textview),
		vrt->textwrap ? GTK_WRAP_WORD : GTK_WRAP_NONE);
	gint char_width, char_height;
	e2_widget_set_font (vrt->textview, e2_option_str_get ("output-font"),
				 &char_width, &char_height);
	// add the text
	_e2_view_dialog_refresh (vrt->textview, file);

	//action area is apparently an hbox, so we need to
	//put the search panel in the vbox, or else repeat all the e2
	//button-add functions to make them work on a custom container ...
	//create search panel
	vrt->panel = e2_widget_add_box (GTK_DIALOG (vrt->dialog)->vbox,
		FALSE, 0, TRUE, FALSE, 0);
	GtkWidget *hbox = e2_widget_add_box (vrt->panel, FALSE, 0,
		FALSE, FALSE, E2_PADDING);
	//add search-string combo
	vrt->combo = e2_combobox_add (hbox,
		FALSE, 0, _e2_view_dialog_search_activate_cb, vrt, 
		E2_COMBOBOX_HAS_ENTRY | E2_COMBOBOX_FOCUS_ON_CHANGE);
	gtk_widget_set_size_request (vrt->combo, 300, -1);
	e2_combobox_append_history (vrt->combo, history);
	_e2_view_dialog_update_search_combo (vrt);

	GtkWidget *btn = e2_button_add_toggle (hbox, _("case sensitive"),
		FALSE, FALSE, 0, _e2_view_dialog_toggled,
		&vrt->case_sensitive, FALSE);
	btn = e2_button_add_toggle (hbox, _("search backward"),
		FALSE, FALSE, 0, _e2_view_dialog_toggled,
		&vrt->search_backward, FALSE);
	btn = e2_button_add_toggle (hbox, _("wrap around"),
		FALSE, FALSE, 0, _e2_view_dialog_toggled,
		&vrt->searchwrap, FALSE);

	gtk_widget_hide (vrt->panel);

	//now add search and/or view buttons to the action-area
	vrt->info_label = e2_widget_add_label (GTK_DIALOG (vrt->dialog)->action_area,
	"", 0, FALSE, 0);
	gtk_widget_hide (vrt->info_label);

	GtkWidget *wrap_button = _e2_dialog_add_check_button
		(vrt->dialog, TRUE, _("_Wrap"), E2_RESPONSE_USER1);
	g_object_set_data (G_OBJECT (vrt->dialog), "wrap", wrap_button);
	vrt->hidebtn = e2_dialog_add_undefined_button
		(vrt->dialog, GTK_STOCK_ZOOM_FIT, _("_Hide"), E2_RESPONSE_USER3);
	//this one not seen yet
	gtk_widget_hide (vrt->hidebtn);
	vrt->nextbtn = e2_dialog_add_undefined_button
		(vrt->dialog, GTK_STOCK_REDO, _("_Next"), E2_RESPONSE_USER2);
	gtk_widget_hide (vrt->nextbtn);
	e2_dialog_add_undefined_button 
		(vrt->dialog, GTK_STOCK_FIND, _("_Find"), E2_RESPONSE_FIND);
	vrt->refreshbtn =
		e2_dialog_add_defined_button (vrt->dialog, &E2_BUTTON_REFRESH);
	e2_dialog_add_defined_button (vrt->dialog, &E2_BUTTON_CLOSE);
//	e2_dialog_set_responses (rt->dialog, E2_RESPONSE_FIND, GTK_RESPONSE_CLOSE);
	e2_dialog_set_negative_response (vrt->dialog, GTK_RESPONSE_CLOSE);
	gtk_dialog_set_default_response (GTK_DIALOG (vrt->dialog), E2_RESPONSE_FIND);

//	GtkWidget *menu = e2_option_create_menu (NULL,
//		opt_show_last, toggled_cb1, rt,
//		opt_case_sensitive, toggled_cb2, rt,
//		opt_view_wrap, view_wrap_cb, rt,
//		NULL);
//	e2_menu_connect (menu,FALSE);
//	e2_dialog_attach_menu (rt->dialog, menu);

	//init search runtime data
	vrt->history = e2_list_copy_with_data (history);
	vrt->textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (vrt->textview));
	vrt->show_last = _e2_option_bool_get (opt_show_last);

	gtk_window_resize (GTK_WINDOW (vrt->dialog), char_width * vrt->window_width,
				char_height * vrt->window_height);

	return vrt->dialog;
}
  /*******************/
 /***** actions *****/
/******************/
/**
 * @brief open file for viewing, at a location where there is a specified string
 * @param filename NULL always
 * @param arg string with filepath and an argument that is a text string to find in the file
 * @return
 **/
void e2_view_dialog_openat (gchar *filename, gchar *arg)
{
	gchar *locator = strchr (arg, ' ');	//always ascii ' ', don't need g_utf8_strchr()
	if (locator == NULL)
		return;	//error msg ??
	*locator = '\0';
	while (*(++locator) == ' ') {}	//FIXME utf, not ascii

	GtkWidget *dialog = _e2_view_dialog_create (arg);
	if (dialog == NULL)
		return;	//error msg ??
	//find the locator and open there
	GtkTextIter iter, start, end;
	gtk_text_buffer_get_start_iter (vrt->textbuffer, &iter);
	if (gtk_text_iter_forward_search (&iter, locator, 0, &start, &end, NULL))
	{	//need to add a mark to get the scroll to work properly
		GtkTextMark *mark = gtk_text_buffer_create_mark
			(vrt->textbuffer, NULL, &start, FALSE);
		gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW (vrt->textview), mark,
			0.0, TRUE, 0.0, 0.1);
		//to support incremental searching, move the cursor
		//to the start of the located string
		gtk_text_buffer_place_cursor (vrt->textbuffer, &start);
	}
	e2_dialog_show (dialog, app.main_window, E2_DIALOG_DONT_SHOW_ALL,
		NULL);
	gtk_window_present (GTK_WINDOW (dialog));
}
  /******************/
 /***** public *****/
/******************/

void e2_view_dialog_show (gchar *filename)
{
	GtkWidget *dialog = _e2_view_dialog_create (filename);
	e2_dialog_show (dialog, app.main_window, E2_DIALOG_DONT_SHOW_ALL, NULL);
	gtk_window_present (GTK_WINDOW (dialog));
}

void e2_view_dialog_actions_register (void)
{
	gchar *action_name = g_strconcat(_A(3),".",_A(81),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		e2_view_dialog_show, NULL, FALSE);
	action_name = g_strconcat(_A(5),".",_A(82),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		e2_view_dialog_openat, NULL, TRUE);
}

void e2_view_dialog_options_register (void)
{
	e2_cache_list_register ("search-history", &history);

//CHECKME = strdup's for group, to support cleanups
	gchar *group_name = g_strconcat(_C(9),".",_C(40),":",_C(24),NULL); //_("dialogs.view:misc"
	opt_view_wrap = e2_option_bool_register ("dialog-view-wrap",
		group_name, _("wrap text"),
		_("This causes the view window to open with text-wrapping enabled"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);
	opt_view_width = e2_option_int_register ("dialog-view-width",
		group_name, _("window width"),
		_("The view window will default to showing this many characters per line (but the the displayed buttons may make it wider than this)")
		, NULL, 84, 20, 1000,
		E2_OPTION_FLAG_ADVANCED);
	opt_view_height = e2_option_int_register ("dialog-view-height",
		group_name, _("window height"),
		_("The view window will default to showing this many lines of text"), NULL, 30, 10, 1000,
		E2_OPTION_FLAG_ADVANCED);
	opt_show_last = e2_option_bool_register ("dialog-search-show-last",
		group_name, _("show last search string"),
		_("This shows the last search-string in the entry field, after displaying the search options or when commencing a new search"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
}
